import cv2
import numpy as np
import onnxruntime as ort

from ..utils import download_file
from .detect import Model
from .ocr import TextSystem


def iou(box1, box2):
    tl1, wh1, br1 = [box1[0], box1[1]], [box1[2], box1[3]], [box1[0] + box1[2], box1[1] + box1[3]]
    tl2, wh2, br2 = [box2[0], box2[1]], [box2[2], box2[3]], [box2[0] + box2[2], box2[1] + box2[3]]
    intersection_area = np.prod(np.maximum(np.minimum(br1, br2) - np.maximum(tl1, tl2), 0))
    union_area = np.prod(wh1) + np.prod(wh2) - intersection_area;
    return intersection_area / union_area


# TODO: Replace with non_maximum_suppression from ops
def nms(objects, iou_threshold = 0.5):
    results = []
    objects.sort(key=lambda obj: obj['score'], reverse=True)
    for object in objects:
        non_overlap = True
        for result in results:
            if iou(object['box'], result['box']) >= iou_threshold:
                non_overlap = False
                break
        if non_overlap:
            results.append(object)
    return results


class LicensePlateDetector():
    def __init__(self, threshold = 0.5, iou_threshold = 0.1, out_size = 300):
        self.out_size = out_size
        self.threshold = threshold
        self.iou_threshold = iou_threshold
        lpd_src = download_file('multiple/models/lpd.onnx')
        self.session = ort.InferenceSession(lpd_src)
    
    def detect(self, img, net_stride = 2**4):
        height, width = img.shape[:2]
        factor = min(288 * max(width, height) / min(width, height), 608) / min(width, height)
        w, h = int(factor * width), int(factor * height)
        w = w // net_stride * net_stride
        h = h // net_stride * net_stride

        rimg = cv2.resize(img, (w, h)).astype(np.float32) / 255
        T = np.expand_dims(rimg, axis=0)
		
        input_name = self.session.get_inputs()[0].name
        pred_onnx = self.session.run(None, {input_name: T})[0]
        Y = np.squeeze(pred_onnx)
        
        objects = []
        probs = Y[...,0]
        affines = Y[...,2:]
        side = ((208 + 40) / 2) / net_stride # 7.75
        yy, xx = np.where(probs > self.threshold)
        for y, x in zip(yy, xx):
            prob = probs[y, x]
            A = affines[y, x].reshape(2, 3)
            A[0, 0], A[1, 1] = max(A[0, 0], 0), max(A[1, 1], 0)
            base = np.matrix([[-0.5, -0.5, 1.],
                              [0.5, -0.5, 1.],
                              [0.5, 0.5, 1.],
                              [-0.5, 0.5, 1.]]).T
            pts = (side * np.dot(A, base).T + np.array([x + .5, y + .5])) * net_stride
            pts = np.array(pts / np.array([w, h])) * np.array([width, height])
            pt1, pt2 = pts.min(axis=0), pts.max(axis=0)
            box = [round(pt1[0]), round(pt1[1]), round(pt2[0] - pt1[0]), round(pt2[1] - pt1[1])]
            objects.append({'box': box, 'score': float(prob), 'points': pts.astype(np.int32)})
        results = nms(objects, self.iou_threshold)
        return results


class PlateDetector():
    def __init__(self):
        model_uri = 'multiple/models/alpd-seg.onnx'
        self.detection = Model(model_uri)

    def detect(self, img):
        return self.detection.run(img, approx=0.02)


def extract_plate(img, points, out_size, offset=30):
    w1, w2 = np.linalg.norm(points[0] - points[1]), np.linalg.norm(points[2] - points[3])
    h1, h2 = np.linalg.norm(points[0] - points[3]), np.linalg.norm(points[1] - points[2])

    w, h = max(w1, w2), max(h1, h2)
    max_wh = max(w, h)

    w, h = int(w / max_wh * out_size), int(h / max_wh * out_size)
    x1, y1, x2, y2 = offset, offset, w + offset, h + offset
    t_points = np.array([[x1, y1], [x2, y1], [x2, y2], [x1, y2]])
    
    size = (w + 2 * offset, h + 2 * offset)
    H, _ = cv2.findHomography(points, t_points)
    return cv2.warpPerspective(img, H, size, borderValue=0.0)


class PlateRecognizer():
    def __init__(self):
        self.license_plate = LicensePlateDetector(threshold=0.85, iou_threshold=0.15)
        self.text_system = TextSystem()
        self.out_size = 300

    def recognize(self, img):
        results = self.license_plate.detect(img)
        for result in results:
            points = result['points']
            img_lp = extract_plate(img, points, self.out_size)
            outputs = self.text_system(img_lp)
            result['lines'] = outputs
        results = [result for result in results if len(result['lines'])]
        for result in results:
            result['label'] = '\n'.join([line.get('text', '') for line in result['lines']])
            del result['score']
        return results



import os
import queue
import threading
import logging
from types import SimpleNamespace

from abraia.utils import download_file, load_json

from .toolbox import (
    MAX_INPUT_QUEUE_SIZE,
    MAX_OUTPUT_QUEUE_SIZE,
    ModelInference,
    get_labels,
    default_preprocess
)
from ..utils.video import VideoInput, VideoDisplay

from ..inference.tracker import TrackletHistory, Tracker
from ..utils.draw import render_results

logger = logging.getLogger(__name__)

DEFAULT_OPTIONS = {
    "input": 0,
    "hef_path": None,
    "task": "detect",
    "labels": None,
    "batch_size": 1,
    "score_threshold": 0.25,
    "model_type": "v5",
    "track": True,
    "draw_trail": False,
    "frame_rate": None,
    "camera_resolution": None,
    "video_unpaced": False,
    "dest": None
}

CONFIG_DATA = {
    "tracker": {
        "track_thresh": 0.1,
        "track_buffer": 30,
        "match_thresh": 0.9
    }
}


def inference_result_handler(original_frame, detections, tracker=None, tracklet_history=None):
    """
    Processes inference results and draw detections (with optional tracking).

    Args:
        original_frame (np.ndarray): Original image frame.
        detections (list): Extracted detections.
        tracker (Tracker, optional): Tracker instance.
        tracklet_history (TrackletHistory, optional): Tracklet history instance.

    Returns:
        np.ndarray: Frame with detections or tracks drawn.
    """
    if tracker:
        detections = tracker.update(detections)
        if tracklet_history:
            tracklet_history.update(detections)
    return render_results(original_frame, detections)


def run_inference_pipeline(
    model_inference: ModelInference,
    input_data: VideoInput,
    visualizer: VideoDisplay,
    tracker: Tracker = None,
    tracklet_history: TrackletHistory = None,
) -> None:
    """
    Initialize queues, inference instance, and run the pipeline.
    """
    input_queue = queue.Queue(MAX_INPUT_QUEUE_SIZE)
    output_queue = queue.Queue(MAX_OUTPUT_QUEUE_SIZE)

    height, width, _ = model_inference.get_input_shape()

    try:
        preprocess_thread = threading.Thread(
            target=input_data.preprocess,
            args=(input_queue, lambda frame: default_preprocess(frame, width, height)),
            name="preprocess-thread",
        )

        inference_thread = threading.Thread(
            target=model_inference.infer,
            args=(input_queue, output_queue, input_data.stop_event),
            name="inference-thread",
        )

        preprocess_thread.start()
        inference_thread.start()

        visualizer.visualize(
            output_queue,
            inference_result_handler,
            is_capture=input_data.has_capture,
            tracker=tracker,
            tracklet_history=tracklet_history,
        )
    finally:
        input_data.stop_event.set()
        preprocess_thread.join()
        inference_thread.join()

    logger.info(visualizer.frame_rate_summary())
    logger.info("Processing completed successfully.")
    if visualizer.dest:
        logger.info(f"Saved outputs to '{visualizer.dest}'.")


def main(**kwargs) -> None:
    """
    Main entry point for the object detection application.

    Args:
        **kwargs: Programmatic arguments to override defaults.

    Example:
        from abraia.hailo import object_detection
        object_detection.main(input='video.mp4', track=True)
    """
    options = DEFAULT_OPTIONS.copy()
    options.update(kwargs)
    args = SimpleNamespace(**options)
    logging.basicConfig(level=logging.INFO)

    try: 
        model_uri = args.hef_path
        config_uri = f"{os.path.splitext(model_uri)[0]}.json"
        config = load_json(download_file(config_uri))
        hef_path = download_file(model_uri)
        labels = config['classes']
        task = config['task']
    except:
        hef_path = args.hef_path
        task = args.task
        labels = get_labels(args.labels)

    stop_event = threading.Event()

    input_data = VideoInput(
        input_src=args.input,
        batch_size=args.batch_size,
        resolution=args.camera_resolution,
        frame_rate=args.frame_rate,
        video_unpaced=args.video_unpaced,
        stop_event=stop_event
    )

    visualizer = VideoDisplay(
        dest=args.dest,
        source_fps=input_data.source_fps,
        stop_event=stop_event
    )

    model_inference = ModelInference(
        hef_path, task, labels,
        batch_size=input_data.batch_size,
        score_threshold=args.score_threshold,
        model_type=args.model_type
    )

    tracker = None
    tracklet_history = None
    if args.track:
        tracker_config = CONFIG_DATA.get("tracker", {})
        tracker = Tracker(
            track_thresh=tracker_config.get('track_thresh', 0.1),
            track_buffer=tracker_config.get('track_buffer', 30),
            match_thresh=tracker_config.get('match_thresh', 0.9),
            frame_rate=input_data.source_fps or 30.0
        )
        if args.draw_trail:
            tracklet_history = TrackletHistory()

    run_inference_pipeline(
        model_inference=model_inference,
        input_data=input_data,
        visualizer=visualizer,
        tracker=tracker,
        tracklet_history=tracklet_history,
    )


if __name__ == "__main__":
    main()



import os
from os import path
import numpy as np
from PIL import Image

from ..client import Abraia
from ..utils import temporal_src, get_type


def load_tiff(src):
    """Load a TIFF image from the given source"""
    import tifffile
    return tifffile.imread(src)


def save_tiff(src, img):
    """Save a TIFF image to the given source"""
    import tifffile
    tifffile.imwrite(src, img)


def decode_mosaic(img, size=(4, 4)):
    """Decode a mosaic image into its spectral bands"""
    r, c = size
    return np.dstack([img[(k % r)::r, (k // c)::c] for k in range(r * c)])


def load_mat(src):
    """Load a MATLAB .mat file from the given source"""
    import scipy
    mat = scipy.io.loadmat(src)
    for key, value in mat.items():
        if type(value) == np.ndarray:
            return value
    return mat


def save_mat(src, img):
    """Save an image to a MATLAB .mat file at the given source"""
    import scipy
    scipy.io.savemat(src, {'raw': img})


def load_image(src):
    if src.lower().endswith('.mat'):
        return load_mat(src)
    if get_type(src) == 'image/tiff':
        return decode_mosaic(load_tiff(src))
    return np.asarray(Image.open(src))


def save_image(src, img):
    if src.lower().endswith('.mat'):
        save_mat(src, img)
    elif get_type(src) == 'image/tiff':
        save_tiff(src, img)
    else:
        Image.fromarray(img).save(src)


def principal_components(img, n_components=3, spectrum=False):
    """Calculate principal components of the image"""
    from sklearn.decomposition import PCA
    h, w, d = img.shape
    X = img.reshape((h * w), d)
    pca = PCA(n_components=n_components, whiten=True)
    bands = np.squeeze(pca.fit_transform(X).reshape(h, w, n_components))
    if spectrum:
        return bands, pca.components_
    return bands


def normalize(img):
    """Normalize the image to the range [0, 1]"""
    min, max = np.amin(img), np.amax(img)
    return (img - min) / (max - min)


def calculate_gray(img):
    """Calculate the grayscale image by PCA method"""
    pc_img = principal_components(img, n_components=1)
    return np.uint8(255 * normalize(np.squeeze(pc_img)))


def create_visible(src):
    img = load_image(src)
    gray = calculate_gray(img / 65535)
    save_image(f"{os.path.splitext(src)[0]}.png", gray)


def random(img, n_bands=6, indexes=False):
    """Return a list of random bands"""
    bands = []
    indexes = []
    for i in range(n_bands):
        q = np.random.randint(img.shape[2])
        indexes.append(q)
        bands.append(img[:, :, q])
    if indexes:
        return bands, indexes
    return bands


def resize(img, size):
    """Resize the image to the given size (w, h)"""
    return np.array(Image.fromarray(img).resize(size, resample=Image.LANCZOS))


def resample(img, n_samples=32):
    """Resample the number of spectral bands (n_samples)"""
    from sklearn.utils import resample
    h, w, d = img.shape
    X = img.reshape((h * w), d)
    r = resample(np.transpose(X), n_samples=n_samples)
    return np.transpose(r).reshape(h, w, n_samples)


try:
    import spectral
    spectral.settings.envi_support_nonlowercase_params = True
except ImportError:
    print('Install the spectral package to work with envi files')


def rgb(img, bands=None):
    """Return the RGB image from the selected bands (R, G, B)"""
    from spectral import get_rgb
    return get_rgb(img, bands=bands)


def ndvi(img, red_band, nir_band):
    """Return the NDVI image from the specified read and nir bands"""
    from spectral import ndvi
    return ndvi(img, red_band, nir_band)


class Multiple(Abraia):
    def __init__(self, folder=''):
        super(Multiple, self).__init__()

    def load_header(self, path):
        dest = temporal_src(path)
        self.download_file(path, dest, cache=True)
        return spectral.io.envi.read_envi_header(dest)

    def load_metadata(self, path):
        if path.lower().endswith('.hdr'):
            return self.load_header(path)
        return super(Multiple, self).load_metadata(path)

    def load_envi(self, path):
        dest = temporal_src(path)
        self.download_file(path, dest, cache=True)
        raw_path = f"{path.split('.')[0]}.raw"
        raw_dest = temporal_src(raw_path)
        self.download_file(raw_path, raw_dest, cache=True)
        return np.array(spectral.io.envi.open(dest, raw_dest)[:, :, :])

    def load_image(self, path):
        if path.lower().endswith('.hdr'):
            return self.load_envi(path)
        dest = temporal_src(path)
        return load_image(self.download_file(path, dest, cache=True))

    def save_envi(self, path, img, metadata={}):
        src = temporal_src(path)
        spectral.io.envi.save_image(src, img, metadata=metadata, force=True)
        self.upload_file(f"{src.split('.')[0]}.img", f"{path.split('.')[0]}.raw")
        return self.upload_file(src, path)

    def save_image(self, path, img, metadata={}):
        src = temporal_src(path)
        if path.lower().endswith('.hdr'):
            return self.save_envi(path, img, metadata)
        else:
            save_image(src, img)
        return self.upload_file(src, path)
    


import os
import sys
import cv2
import time
import queue
import logging
import threading
import numpy as np

from pathlib import Path
from typing import Dict, Generator, List, Optional, Tuple, Callable, Any

from .draw import (
    render_resolution,
    render_status,
    calculate_optimal_thickness,
    calculate_optimal_text_scale,
)

logger = logging.getLogger(__name__)


def make_dirs(dest):
    """Create directory if it doesn't exist."""
    dirname = os.path.dirname(dest)
    if dirname:
        os.makedirs(dirname, exist_ok=True)


VIDEO_SUFFIXES = (".mp4", ".avi", ".mov", ".mkv")
IMAGE_EXTENSIONS: Tuple[str, ...] = (".jpg", ".jpeg", ".png", ".bmp")
CAMERA_RESOLUTION_MAP: Dict[str, Tuple[int, int]] = {
    "sd": (640, 480), "hd": (1280, 720), "fhd": (1920, 1080)
}


def is_raspberry_pi() -> bool:
    """Check if the current host is a Raspberry Pi."""
    try:
        with open('/proc/device-tree/model', 'r') as f:
            return "Raspberry Pi" in f.read()
    except Exception:
        return False


class PiCamera2CaptureAdapter:
    """Adapter that makes Picamera2 behave like cv2.VideoCapture."""

    def __init__(self, picam2):
        self.picam2 = picam2
        self._opened = True
        self._io_lock = threading.Lock()

    def isOpened(self):
        return self._opened

    def read(self):
        if not self._opened:
            return False, None

        # prevent stop/close while capturing
        with self._io_lock:
            if not self._opened: # re-check after taking lock
                return False, None
            frame = self.picam2.capture_array()

        if frame is None:
            return False, None
        return True, frame

    def get(self, prop_id: int) -> float:
        if prop_id in (cv2.CAP_PROP_FRAME_WIDTH, cv2.CAP_PROP_FRAME_HEIGHT):
            try:
                cfg = self.picam2.camera_configuration()
                size = cfg.get("main", {}).get("size", None)
                if size and len(size) == 2:
                    w, h = int(size[0]), int(size[1])
                    return float(w if prop_id == cv2.CAP_PROP_FRAME_WIDTH else h)
            except Exception:
                pass
            return 0.0
        if prop_id == cv2.CAP_PROP_FPS:
            return 30.0
        return None

    def release(self):
        # stop new reads ASAP
        self._opened = False
        # wait if a read() is currently inside capture_array()
        with self._io_lock:
            try:
                self.picam2.stop()
            except Exception:
                pass
            try:
                self.picam2.close()
            except Exception:
                pass


def open_cv_capture(src: Any, source_type: str, resolution=(1280, 720), fps=30) -> Any:
    """Open an OpenCV-based capture source."""
    if source_type == "video" and not os.path.exists(src):
        logger.error(f"File not found: {src}")
        sys.exit(1)
    cap = cv2.VideoCapture(int(src) if source_type == "usb_camera" else src)
    if source_type == "usb_camera":
        cap.set(cv2.CAP_PROP_FRAME_WIDTH, resolution[0])
        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, resolution[1])
        cap.set(cv2.CAP_PROP_FPS, fps)
    if not cap.isOpened():
        logger.error(f"Failed to open {source_type} source: {src}")
        sys.exit(1)
    logger.info(f"Using {source_type} input: {src}")
    return cap


def open_rpi_camera(resolution=(1280, 720), fps=30) -> Optional[Any]:
    """Open Raspberry Pi camera using Picamera2."""
    try:
        from picamera2 import Picamera2
    except Exception as e:
        logger.error(f"Picamera2 not available: {e}")
        return None
    try:
        picam2 = Picamera2()
        config = picam2.create_video_configuration(main={"size": resolution, "format": "RGB888"}, controls={"FrameRate": fps})
        picam2.configure(config)
        picam2.start()
        return PiCamera2CaptureAdapter(picam2)
    except Exception as e:
        logger.error(f"Failed to open RPi camera: {e}")
        try:
            picam2.stop()
        except Exception:
            pass
        try:
            picam2.close()
        except Exception:
            pass
        return None


def is_stream_url(src: str) -> bool:
    """Return True if the input looks like a supported network stream URL."""
    src_lower = src.lower()
    return (src_lower.startswith("rtsp://") or src_lower.startswith("http://") or src_lower.startswith("https://"))


def is_video(src: str) -> bool:
    """Return True if the input is a video file."""
    return os.path.isfile(src) and src.lower().endswith(VIDEO_SUFFIXES)


def is_image(src: str) -> bool:
    """Return True if the input is an image file."""
    return os.path.isfile(src) and src.lower().endswith(IMAGE_EXTENSIONS)


def load_images_opencv(images_path: str) -> List[np.ndarray]:
    path = Path(images_path)
    def read_rgb(p: Path):
        img = cv2.imread(str(p))
        if img is not None:
            return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return None
    if path.is_file() and path.suffix.lower() in IMAGE_EXTENSIONS:
        img = read_rgb(path)
        return [img] if img is not None else []
    elif path.is_dir():
        images = [read_rgb(img) for img in path.glob("*") if img.suffix.lower() in IMAGE_EXTENSIONS]
        return [img for img in images if img is not None]
    return []


def get_input_type(src: Any) -> str:
    """Determine the type of input source."""
    src_str = str(src)
    if src_str.isdigit():
        return "rpi_camera" if is_raspberry_pi() else "usb_camera"
    if is_stream_url(src_str):
        return "stream"
    if os.path.exists(src_str):
        return "video" if is_video(src_str) else "images"


class VideoInput:
    def __init__(
        self,
        input_src: str,
        batch_size: int = 1,
        resolution: Optional[str] = None,
        frame_rate: Optional[float] = None,
        video_unpaced: bool = False,
        stop_event: Optional[threading.Event] = None,
    ):
        self.batch_size = batch_size
        self.resolution = resolution
        self.frame_rate = frame_rate
        self.video_unpaced = video_unpaced
        self.stop_event = stop_event or threading.Event()

        self.cap = None
        self.images = None

        self.input_type = get_input_type(input_src)
        if not self.input_type:
            logger.error(f"Invalid input '{input_src}'.")
            sys.exit(1)

        self.width = None
        self.height = None
        self.source_fps = None
        if self.input_type == "images":
            self.images = load_images_opencv(input_src)
        else:
            width, height = 1280, 720
            if self.resolution in CAMERA_RESOLUTION_MAP:
                width, height = CAMERA_RESOLUTION_MAP[self.resolution]
            if self.input_type == "rpi_camera":
                self.cap = open_rpi_camera(resolution=(width, height))
                self.source_fps = 30
            else:
                self.cap = open_cv_capture(input_src, self.input_type, resolution=(width, height))
            if self.cap is not None:
                self.source_fps = self.cap.get(cv2.CAP_PROP_FPS)
                self.width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH) or 0)
                self.height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT) or 0)

    @property
    def has_capture(self) -> bool:
        return self.cap is not None

    @property
    def has_images(self) -> bool:
        return self.images is not None and len(self.images) > 0

    def _generate_frames(self) -> Generator[np.ndarray, None, None]:
        if self.has_images:
            yield from self.images
            return

        is_camera = self.input_type in ("usb_camera", "rpi_camera", "stream")
        is_video = self.input_type == "video"
        target_fps = self.frame_rate or 0
        source_fps = self.source_fps or 0
        
        should_drop = target_fps > 0 and (source_fps == 0 or target_fps < source_fps)
        should_pace = is_video and not self.video_unpaced
        
        keep_period = 1.0 / target_fps if should_drop and is_camera else 0
        video_keep_period_ms = 1000.0 / target_fps if should_drop and is_video and should_pace else 0
        
        next_keep_timestamp = time.monotonic()
        video_start_ms, wall_start_time = None, None
        next_keep_video_ms = None

        while not self.stop_event.is_set():
            ret, frame_bgr = self.cap.read()
            if not ret: break
            
            current_pos_ms = float(self.cap.get(cv2.CAP_PROP_POS_MSEC) or 0.0)
            if should_pace:
                if video_start_ms is None:
                    video_start_ms, wall_start_time = current_pos_ms, time.monotonic()
                if video_keep_period_ms:
                    if next_keep_video_ms is None: next_keep_video_ms = current_pos_ms
                    if current_pos_ms + 1e-3 < next_keep_video_ms: continue
                    while current_pos_ms + 1e-3 >= next_keep_video_ms: next_keep_video_ms += video_keep_period_ms
                desired_wall_time = wall_start_time + (current_pos_ms - video_start_ms) / 1000.0
                if time.monotonic() < desired_wall_time:
                    time.sleep(max(0, desired_wall_time - time.monotonic()))
            
            if keep_period:
                if time.monotonic() < next_keep_timestamp: continue
                next_keep_timestamp += keep_period
            
            yield cv2.cvtColor(frame_bgr, cv2.COLOR_BGR2RGB)
        self.cap.release()

    def preprocess(self, input_queue: queue.Queue, preprocess_fn: Callable[[np.ndarray], np.ndarray]) -> None:
        raw_frames, processed_frames = [], []
        for frame in self._generate_frames():
            raw_frames.append(frame)
            processed_frames.append(preprocess_fn(frame))
            if len(raw_frames) >= self.batch_size:
                input_queue.put((raw_frames, processed_frames))
                raw_frames, processed_frames = [], []
        if raw_frames:
            input_queue.put((raw_frames, processed_frames))
        input_queue.put(None)


class VideoDisplay:
    def __init__(self, dest: Optional[str] = None, source_fps: Optional[float] = None, stop_event: Optional[threading.Event] = None):
        self.dest = dest
        self.source_fps = source_fps
        self.stop_event = stop_event or threading.Event()

        self._count = 0
        self._start_time = None

        self.video_writer = None
        self.image_index = 0
        self.window_name = "Output"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.video_writer is not None:
            self.video_writer.release()
        cv2.destroyAllWindows()

    def show(self, frame: np.ndarray, fps: float, is_capture: bool = True) -> bool:
        if not hasattr(self, 'thickness'):
            self.thickness = calculate_optimal_thickness(frame.shape[:2])
            self.text_scale = calculate_optimal_text_scale(frame.shape[:2])

        render_status(frame, fps, thickness=self.thickness, text_scale=self.text_scale)
        render_resolution(frame, thickness=self.thickness, text_scale=self.text_scale)

        output_bgr_frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        if self.dest:
            if is_capture:
                if self.video_writer is None:
                    make_dirs(self.dest)
                    height, width = frame.shape[:2]
                    fourcc = cv2.VideoWriter_fourcc(*"XVID")
                    self.video_writer = cv2.VideoWriter(self.dest, fourcc, self.source_fps, (width, height))
                if self.video_writer:
                    self.video_writer.write(output_bgr_frame)
            else:
                path = Path(self.dest)
                out_path = path.parent / f"{path.stem}_{self.image_index}{path.suffix}"
                out_path.parent.mkdir(parents=True, exist_ok=True)
                cv2.imwrite(str(out_path), output_bgr_frame)
                self.image_index += 1
        cv2.imshow(self.window_name, output_bgr_frame)
        if (cv2.waitKey(1) & 0xFF) == ord("q"):
            return False
        return True

    def start(self):
        self._start_time = time.time()

    def increment(self, n: int = 1):
        self._count += n

    @property
    def count(self) -> int:
        return self._count

    @property
    def elapsed(self) -> float:
        if self._start_time is None:
            return 0.0
        return time.time() - self._start_time

    @property
    def fps(self) -> float:
        elapsed = self.elapsed
        return self._count / elapsed if elapsed > 0 else 0.0

    def frame_rate_summary(self) -> str:
        return f"Processed {self.count} frames at {self.fps:.2f} FPS, Total time: {self.elapsed:.2f} seconds"

    def visualize(self, output_queue: queue.Queue, callback: Callable, is_capture: bool = True, **kwargs) -> None:
        self.start()
        with self:
            while True:
                try:
                    result = output_queue.get()
                    if result is None:
                        break
                    original_frame, inference_result = result
                    if self.stop_event.is_set():
                        continue
                    frame_with_detections = callback(original_frame, inference_result, **kwargs)
                    self.increment()
                    if not self.show(frame_with_detections, self.fps, is_capture=is_capture):
                        self.stop_event.set()
                finally:
                    output_queue.task_done()
        self.stop_event.set()


class Video:
    def __init__(self, src=0, resolution=(1920, 1080), fps=30, dest=None):
        self.out = None
        self.quit = False
        self.win_name = ''
        self.picam2 = None
        if src == 0 and is_raspberry_pi():
            from picamera2 import Picamera2
            self.picam2 = Picamera2()
            self.picam2.configure(self.picam2.create_video_configuration(
                main={"format": 'RGB888', "size": resolution},
                controls={"FrameRate": fps}))
            self.picam2.start()
            self.fps = fps
            self.width, self.height = resolution
            self.frames = 0
            self.duration = 0
        else:
            self.cap = cv2.VideoCapture(src)
            if isinstance(src, int):
                self.cap.set(cv2.CAP_PROP_FPS, fps)
                self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, resolution[0])
                self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, resolution[1])
            self.fps = self.cap.get(cv2.CAP_PROP_FPS)
            self.frames = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
            self.width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            self.height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            self.duration = round(self.frames / self.fps, 3) if self.fps > 0 else 0
        self.frame_rate = self.fps
        if dest:
            make_dirs(dest)
            # fourcc = cv2.VideoWriter_fourcc(*'avc1')
            fourcc = cv2.VideoWriter_fourcc(*"XVID")
            self.out = cv2.VideoWriter(dest, fourcc, self.fps, (self.width, self.height))
        self.t0 = time.time()

    def __len__(self):
        return self.frames

    def __iter__(self):
        if self.picam2:
            while not self.quit:
                frame = self.picam2.capture_array()
                if frame is None:
                    break
                yield cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            self.picam2.stop()
        else:
            while self.cap.isOpened():
                ret, frame = self.cap.read()
                if ret is False or frame is None or self.quit:
                    break
                yield cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            self.cap.release()
        if self.out:
            self.out.release()
        if self.win_name:
            cv2.destroyWindow(self.win_name)
            cv2.waitKey(1)

    def get_frame(self, frame_num):
        if self.picam2:
            return None
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, frame_num)
        ret, frame = self.cap.read()
        if ret is False or frame is None:
            return None
        return cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    def write(self, frame):
        #TODO: Remove on next releases
        if self.out:
            self.out.write(cv2.cvtColor(frame, cv2.COLOR_RGB2BGR))
    
    def show(self, frame):
        t1 = time.time()
        render_status(frame, fps=1 / (t1 - self.t0) if t1 > self.t0 else 0)
        render_resolution(frame)
        self.t0 = t1
        out = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        if self.out:
            self.out.write(out)
        if not self.win_name:
            self.win_name = 'Video'
            cv2.namedWindow(self.win_name, cv2.WINDOW_GUI_NORMAL)
        cv2.imshow(self.win_name, out)
        ch = cv2.waitKey(1) & 0xFF
        if (ch == 27 or ch == ord('q')) or cv2.getWindowProperty(self.win_name, cv2.WND_PROP_VISIBLE) < 1:
            self.quit = True


if __name__ == "__main__":
    video = Video()
    for frame in video:
        video.show(frame)



import os
import gzip
import json
import base64
import hashlib
import tempfile
import requests
import mimetypes
import numpy as np
import onnxruntime as ort
import pillow_heif

from tqdm import tqdm
from io import BytesIO
from pathlib import Path
from PIL import Image, ImageOps
from concurrent.futures import ProcessPoolExecutor

from .video import Video
from .sketcher import Sketcher
from .draw import get_color, render_results

pillow_heif.register_heif_opener()

tempdir = tempfile.gettempdir()

API_URL = 'https://api.abraia.me'

HEADERS = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36'}

mimetypes.add_type('image/webp', '.webp')
mimetypes.add_type('image/heic', '.heic')
mimetypes.add_type('image/heif', '.heif')


def get_type(path):
    return mimetypes.guess_type(path)[0] or 'binary/octet-stream'


def md5sum(src):
    hash_md5 = hashlib.md5()
    f = BytesIO(src.getvalue()) if isinstance(src, BytesIO) else open(src, 'rb')
    for chunk in iter(lambda: f.read(4096), b''):
        hash_md5.update(chunk)
    f.close()
    return hash_md5.hexdigest()


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types."""
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def is_url(url):
    return url.startswith('http://') or url.startswith('https://')


def url_path(path):
    return f"{API_URL}/files/{path}"


def list_dir(folder):
    return [os.path.join(folder, f) for f in os.listdir(folder) if os.path.isfile(os.path.join(folder, f))]


def make_dirs(dest):
    """Create directory if it doesn't exist."""
    dirname = os.path.dirname(dest)
    if dirname:
        os.makedirs(dirname, exist_ok=True)


def get_remote_file_size(url: str, timeout: int = 30):
    """Get the size of a remote file via HEAD request."""
    try:
        r = requests.head(url, headers=HEADERS, timeout=timeout, allow_redirects=True)
        length = r.headers.get('Content-Length')
        return int(length) if length else None
    except Exception:
        return None


def temporal_src(path):
    dest = os.path.join(tempdir, path)
    make_dirs(dest)
    return dest


def download_url(url: str, dest: str, chunk_size: int = 8192):
    filename = os.path.basename(dest)
    temp_dest = Path(dest).with_name(filename + '.part')
    temp_dest.parent.mkdir(parents=True, exist_ok=True)
    with requests.get(url, headers=HEADERS, stream=True, allow_redirects=True) as r:
        r.raise_for_status()
        total = int(r.headers.get('content-length', 0))
        with open(temp_dest, 'wb') as f, tqdm(desc=filename, total=total, unit='iB', unit_scale=True, unit_divisor=1024) as bar:
            for chunk in r.iter_content(chunk_size=chunk_size): 
                size = f.write(chunk)
                bar.update(size)
            f.flush()
    temp_dest.rename(dest)


def download_file(path):
    dest = temporal_src(path)
    if not os.path.exists(dest):
        download_url(url_path(path), dest)
    return dest


def load_url(url):
    r = requests.get(url, headers=HEADERS, stream=True, allow_redirects=True)
    return r.raw if r.status_code == 200 else None


def load_json(src, gz=False):
    with gzip.open(src, 'rt') if gz else open(src, 'r') as f:
        return json.load(f)
    

def save_json(data, dest, gz=False):
    make_dirs(dest)
    with gzip.open(dest, 'wt') if gz else open(dest, 'w') as f:    
        f.write(json.dumps(data, cls=NumpyEncoder))
    return dest


def load_text(src, gz=False):
    with gzip.open(src, 'rt') if gz else open(src, 'r') as f:
        return f.read()


def save_text(dest, text, gz=False):
    make_dirs(dest)
    with gzip.open(dest, 'wt') if gz else open(dest, 'w') as f:
        f.write(text)
    return dest


def load_data(src, gz=False):
    with gzip.open(src, 'rb') if gz else open(src, 'rb') as f:
        return f.read()


def save_data(dest, data, gz=False):
    make_dirs(dest)
    with gzip.open(dest, 'wb') if gz else open(dest, 'wb') as f:
        f.write(data)
    return dest


def load_image(src, mode='RGB', max_size=2048):
    im = ImageOps.exif_transpose(Image.open(src).convert(mode)) 
    if im.width > max_size or im.height > max_size:
        im.thumbnail((max_size, max_size), Image.LANCZOS)
    return np.array(im)


def save_image(img, dest):
    make_dirs(dest)
    Image.fromarray(img).save(dest)
    return dest


def show_image(img):
    Image.fromarray(img).show()


def image_base64(img, format='jpeg'):
    im = Image.fromarray(img)
    with BytesIO() as buffer:
        im.save(buffer, format=format)
        encoded = base64.b64encode(buffer.getvalue()).decode()
        return f'data:image/{format};base64,{encoded}'


def get_providers():
    available_providers = ort.get_available_providers()
    providers = ["CUDAExecutionProvider", "CoreMLExecutionProvider", "CPUExecutionProvider"]
    return [provider for provider in available_providers if provider in providers]


def process_map(task, *values, desc='', max_workers=3):
    with ProcessPoolExecutor(max_workers) as exe:
        with tqdm(total=len(values[0]), desc=desc) as pbar:
            for result in exe.map(task, *values):
                pbar.set_postfix_str(result)
                pbar.update(1)


def process_media(src, callback):
    if get_type(str(src)).startswith('image'):
        img = load_image(src)
        out = callback(img)
        show_image(out)
    else:
        video = Video(src)
        for frame in video:
            out = callback(frame)
            video.show(out)




# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cv2
import math
import string
# import pyclipper
import numpy as np
import onnxruntime as ort

from ..utils import download_file


def get_char(character_dict_path, use_space_char=False):
    character_str = ""
    with open(character_dict_path, "rb") as fin:
        lines = fin.readlines()
        for line in lines:
            character_str += line.decode("utf-8").strip("\n").strip("\r\n")
    if use_space_char:
        character_str += " "
    return character_str


class BaseRecLabelDecode():
    """ Convert between text-label and text-index """
    def __init__(self, character_dict_path=None, use_space_char=False):
        self.beg_str = "sos"
        self.end_str = "eos"
        character_str = get_char(character_dict_path, use_space_char)
        dict_character = ['blank'] + list(character_str)
        self.dict = {char: i for i, char in enumerate(dict_character)}
        self.character = dict_character

    def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
        """Convert text-index into text-label."""
        result_list = []
        ignored_tokens = self.get_ignored_tokens()
        batch_size = len(text_index)
        for batch_idx in range(batch_size):
            char_list, conf_list = [], []
            for idx in range(len(text_index[batch_idx])):
                if text_index[batch_idx][idx] in ignored_tokens:
                    continue
                if is_remove_duplicate:
                    if idx > 0 and text_index[batch_idx][idx - 1] == text_index[batch_idx][idx]:
                        continue
                char_list.append(self.character[int(text_index[batch_idx][idx])])
                conf_list.append(text_prob[batch_idx][idx] if text_prob is not None else 1)
            text = ''.join(char_list)
            result_list.append((text, np.mean(conf_list)))
        return result_list

    def get_ignored_tokens(self):
        return [0]  # for ctc blank

    def __call__(self, preds):
        preds_idx, preds_prob = preds.argmax(axis=2), preds.max(axis=2)
        text = self.decode(preds_idx, preds_prob, is_remove_duplicate=True)
        return text


def makeOffsetPoly(poly, offset):
    ccw, offset = np.sign(offset), np.abs(offset)
    def normalizeVec(vec):
        return vec / np.linalg.norm(vec)
    points = []
    num_points = len(poly)
    for curr in range(num_points):
        prev, next = (curr + num_points - 1) % num_points, (curr + 1) % num_points
        vnnX, vnnY = normalizeVec(poly[next] - poly[curr])
        nnnX, nnnY = np.array([vnnY, -vnnX])
        vpnX, vpnY = normalizeVec(poly[curr] - poly[prev])
        npnX, npnY = np.array([vpnY, -vpnX]) * ccw
        bisn = normalizeVec(np.array([nnnX + npnX, nnnY + npnY]) * ccw)
        bislen = offset /  np.sqrt((1 + nnnX * npnX + nnnY * npnY) / 2)
        points.append(poly[curr] + bisn * bislen)
    return np.array(points).round().astype(np.int32)


class DBPostProcess():
    """The post process for Differentiable Binarization (DB)."""

    def __init__(self, thresh=0.3, box_thresh=0.7, max_candidates=1000, unclip_ratio=2.0):
        self.thresh = thresh
        self.box_thresh = box_thresh
        self.max_candidates = max_candidates
        self.unclip_ratio = unclip_ratio
        self.min_size = 3

    def unclip(self, box, unclip_ratio):
        area = cv2.contourArea(box)
        perimeter = cv2.arcLength(box, True)
        distance = round(area * unclip_ratio / perimeter)
        # offset = pyclipper.PyclipperOffset()
        # offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        # expanded = np.array(offset.Execute(distance))
        expanded = makeOffsetPoly(box, distance)
        return expanded.reshape(-1, 1, 2)

    def box_score_fast(self, bitmap, _box):
        """Use bbox mean score as the mean score."""
        h, w = bitmap.shape[:2]
        box = _box.copy()
        xmin = np.clip(np.floor(box[:, 0].min()).astype(np.int32), 0, w - 1)
        ymin = np.clip(np.floor(box[:, 1].min()).astype(np.int32), 0, h - 1)
        xmax = np.clip(np.ceil(box[:, 0].max()).astype(np.int32), 0, w - 1)
        ymax = np.clip(np.ceil(box[:, 1].max()).astype(np.int32), 0, h - 1)
        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
        box = box - np.array([xmin, ymin])
        cv2.fillPoly(mask, box.reshape(1, -1, 2).astype(np.int32), 1)
        return cv2.mean(bitmap[ymin:ymax + 1, xmin:xmax + 1], mask)[0]

    def get_mini_boxes(self, contour):
        bounding_box = cv2.minAreaRect(contour)
        points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])
        index_1, index_4 = (0, 1) if points[1][1] > points[0][1] else (1, 0)
        index_2, index_3 = (2, 3) if points[3][1] > points[2][1] else (3, 2)
        box = [points[index_1], points[index_2], points[index_3], points[index_4]]
        return np.array(box), min(bounding_box[1])

    def boxes_from_bitmap(self, pred, bitmap, dest_size):
        """_bitmap: single map with shape (1, H, W), whose values are binarized as {0, 1}"""
        dest_width, dest_height = dest_size
        height, width = bitmap.shape

        contours, _ = cv2.findContours((bitmap * 255).astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        num_contours = min(len(contours), self.max_candidates)

        boxes, scores = [], []
        for index in range(num_contours):
            contour = contours[index]
            points, sside = self.get_mini_boxes(contour)
            if sside < self.min_size:
                continue
            score = self.box_score_fast(pred, points)
            if self.box_thresh > score:
                continue
            box = self.unclip(points, self.unclip_ratio)
            box, sside = self.get_mini_boxes(box)
            if sside < self.min_size + 2:
                continue
            box[:, 0] = np.clip(np.round(box[:, 0] / width * dest_width), 0, dest_width)
            box[:, 1] = np.clip(np.round(box[:, 1] / height * dest_height), 0, dest_height)
            boxes.append(box)
            scores.append(score)
        return np.array(boxes, dtype=np.int16), scores

    def __call__(self, outs_dict, image_size):
        pred = outs_dict['maps'][:, 0, :, :]
        segmentation = pred > self.thresh
        boxes_batch = []
        for batch_index in range(pred.shape[0]):
            mask = segmentation[batch_index]
            boxes, scores = self.boxes_from_bitmap(pred[batch_index], mask, image_size)
            boxes_batch.append({'points': boxes})
        return boxes_batch


def resize_img(img, limit_side_len):
    """Resize image to a size multiple of 32 which is required by the network."""
    h, w = img.shape[:2]
    ratio = min(limit_side_len / max(h, w), 1)
    resize_h = round(h * ratio / 32) * 32
    resize_w = round(w * ratio / 32) * 32
    img = cv2.resize(img, (resize_w, resize_h))
    ratio_h, ratio_w = resize_h / h, resize_w / w
    return img, (ratio_h, ratio_w)


def str_count(s):
    """Count the number of Chinese characters. A single English character and a single number
    equal to half the length of Chinese characters."""
    count_zh = count_pu = 0
    s_len = len(s)
    en_dg_count = 0
    for c in s:
        if c in string.ascii_letters or c.isdigit() or c.isspace():
            en_dg_count += 1
        elif c.isalpha():
            count_zh += 1
        else:
            count_pu += 1
    return s_len - math.ceil(en_dg_count / 2)


def sorted_boxes(dt_boxes):
    """Sort text boxes in order from top to bottom, left to right
    args:
        dt_boxes(array):detected text boxes with shape [4, 2]
    return:
        sorted boxes(array) with shape [4, 2]
    """
    num_boxes = dt_boxes.shape[0]
    sorted_boxes = sorted(dt_boxes, key=lambda x: (x[0][1], x[0][0]))
    _boxes = list(sorted_boxes)
    for i in range(num_boxes - 1):
        if abs(_boxes[i + 1][0][1] - _boxes[i][0][1]) < 10 and \
            (_boxes[i + 1][0][0] < _boxes[i][0][0]):
            tmp = _boxes[i]
            _boxes[i] = _boxes[i + 1]
            _boxes[i + 1] = tmp
    return _boxes


class TextDetector():
    def __init__(self):
        self.postprocess_op = DBPostProcess(thresh=0.3, box_thresh=0.5, max_candidates=1000, unclip_ratio=1.6)
        det_src = download_file('multiple/models/ocr_det.onnx')
        self.session = ort.InferenceSession(det_src)
        self.input_name = self.session.get_inputs()[0].name
    
    def order_points_clockwise(self, pts):
        """
        reference from: https://github.com/jrosebr1/imutils/blob/master/imutils/perspective.py
        # sort the points based on their x-coordinates
        """
        xSorted = pts[np.argsort(pts[:, 0]), :]
        leftMost, rightMost = xSorted[:2, :], xSorted[2:, :]
        # now, sort the left-most coordinates according to their
        # y-coordinates so we can grab the top-left and bottom-left
        # points, respectively
        tl, bl = leftMost[np.argsort(leftMost[:, 1]), :]
        tr, br = rightMost[np.argsort(rightMost[:, 1]), :]
        rect = np.array([tl, tr, br, bl], dtype="float32")
        return rect

    def clip_det_res(self, points, img_height, img_width):
        for pno in range(points.shape[0]):
            points[pno, 0] = int(min(max(points[pno, 0], 0), img_width - 1))
            points[pno, 1] = int(min(max(points[pno, 1], 0), img_height - 1))
        return points

    def filter_tag_det_res(self, dt_boxes, img_size):
        img_width, img_height = img_size
        dt_boxes_new = []
        for box in dt_boxes:
            box = self.order_points_clockwise(box)
            box = self.clip_det_res(box, img_height, img_width)
            rect_width = int(np.linalg.norm(box[0] - box[1]))
            rect_height = int(np.linalg.norm(box[0] - box[3]))
            if rect_width <= 3 or rect_height <= 3:
                continue
            dt_boxes_new.append(box)
        return np.array(dt_boxes_new)

    def preprocess(self, img):
        img, shape_ratio = resize_img(img, limit_side_len=960)
        img = (img / 255 - np.array([0.485, 0.456, 0.406])) / np.array([0.229, 0.224, 0.225])
        img = img.astype(np.float32).transpose((2, 0, 1))
        return np.expand_dims(img, axis=0)

    def __call__(self, img):
        height, width = img.shape[:2]
        inputs = {self.input_name: self.preprocess(img)}
        outputs = self.session.run(None, inputs)
        post_result = self.postprocess_op({'maps': outputs[0]}, (width, height))
        dt_boxes = post_result[0]['points']
        dt_boxes = self.filter_tag_det_res(dt_boxes, (width, height))
        return dt_boxes


class TextRecognizer():
    def __init__(self):
        self.rec_image_shape = [3, 32, 320]
        self.rec_batch_num = 6
        self.max_text_length = 25
        self.limited_max_width = 1280
        self.limited_min_width = 16

        char_dict_src = download_file('multiple/models/ppocr_keys_v1.txt')
        self.postprocess_op = BaseRecLabelDecode(char_dict_src, use_space_char=True)

        rec_src = download_file('multiple/models/ocr_rec.onnx')
        self.session = ort.InferenceSession(rec_src)
       
    def resize_norm_img(self, img, max_wh_ratio):
        imgC, imgH, imgW = self.rec_image_shape
        max_wh_ratio = max(max_wh_ratio, imgW / imgH)
        imgW = int((32 * max_wh_ratio))
        imgW = max(min(imgW, self.limited_max_width), self.limited_min_width)
        h, w = img.shape[:2]
        ratio_imgH = math.ceil(imgH * w / h)
        ratio_imgH = max(ratio_imgH, self.limited_min_width)
        resized_w = imgW if ratio_imgH > imgW else int(ratio_imgH)
        resized_image = cv2.resize(img, (resized_w, imgH))
        resized_image = (resized_image / 255 - 0.5) / 0.5
        resized_image = resized_image.astype(np.float32).transpose((2, 0, 1))
        padding_im = np.zeros((imgC, imgH, imgW), dtype=np.float32)
        padding_im[:, :, 0:resized_w] = resized_image
        return padding_im

    def __call__(self, img_list):
        img_num = len(img_list)
        width_list = [img.shape[1] / img.shape[0] for img in img_list]
        # Sorting can speed up the recognition process
        indices = np.argsort(np.array(width_list))

        rec_res = [['', 0.0]] * img_num
        batch_num = self.rec_batch_num
        for beg_img_no in range(0, img_num, batch_num):
            end_img_no = min(img_num, beg_img_no + batch_num)
            norm_img_batch = []
            max_wh_ratio = 0
            for ino in range(beg_img_no, end_img_no):
                # h, w = img_list[ino].shape[0:2]
                h, w = img_list[indices[ino]].shape[0:2]
                wh_ratio = w * 1.0 / h
                max_wh_ratio = max(max_wh_ratio, wh_ratio)
            for ino in range(beg_img_no, end_img_no):
                norm_img = self.resize_norm_img(img_list[indices[ino]], max_wh_ratio)
                norm_img = norm_img[np.newaxis, :]
                norm_img_batch.append(norm_img)

            norm_img_batch = np.concatenate(norm_img_batch)
            ort_inputs = {self.session.get_inputs()[0].name: norm_img_batch}
            ort_outs = self.session.run(None, ort_inputs)
            outputs1 = ort_outs[0]
        
        rec_result = self.postprocess_op(outputs1)
        for rno in range(len(rec_result)):
            rec_res[indices[beg_img_no + rno]] = rec_result[rno]
        return rec_res


class TextSystem():
    def __init__(self):
        self.text_detector = TextDetector()
        self.text_recognizer = TextRecognizer()
        self.drop_score = 0.5
        
    def get_rotate_crop_image(self, img, points):
        width = int(max(np.linalg.norm(points[0] - points[1]), np.linalg.norm(points[2] - points[3])))
        height = int(max(np.linalg.norm(points[0] - points[3]), np.linalg.norm(points[1] - points[2])))
        pts_std = np.float32([[0, 0], [width, 0], [width, height], [0, height]])
        M = cv2.getPerspectiveTransform(points, pts_std)
        dst_img = cv2.warpPerspective(img, M, (width, height), borderMode=cv2.BORDER_REPLICATE, flags=cv2.INTER_CUBIC)
        dst_img_height, dst_img_width = dst_img.shape[0:2]
        dst_img = np.rot90(dst_img) if dst_img_height * 1.0 / dst_img_width >= 1.5 else dst_img
        return dst_img

    def __call__(self, img):
        results = []
        dt_boxes = self.text_detector(img)
        if len(dt_boxes):
            dt_boxes = sorted_boxes(dt_boxes)
            img_crop_list = [self.get_rotate_crop_image(img, tmp_box) for tmp_box in dt_boxes]
            rec_res = self.text_recognizer(img_crop_list)
            for box, (text, score) in zip(dt_boxes, rec_res):
                if score >= self.drop_score:
                    results.append({'box': box.astype(np.int32), 'text': text, 'score': float(score)})
        return results 
        


import os
import logging
import numpy as np
import onnxruntime as ort

from PIL import Image
from typing import Union, Iterable, Optional

from .preprocessor import Preprocessor
from .tokenizer import Tokenizer

from ...utils import download_url, download_file


class Clip:
    """
    This class can be utilised to predict the most relevant text snippet, given
    an image, without directly optimizing for the task. This class don't depend on
    `torch` or `torchvision`.
    """

    def __init__(self, batch_size: Optional[int] = None, cache_dir: Optional[str] = 'models'):
        """
        Instantiates the model and required encoding classes.

        Args:
            batch_size: If set, splits the lists in `get_image_embeddings`
                and `get_text_embeddings` into batches of this size before
                passing them to the model. The embeddings are then concatenated
                back together before being returned. This is necessary when
                passing large amounts of data (perhaps ~100 or more).
            cache_dir: If provided, the models will be downloaded to / loaded from this location
        """

        image_model_path = download_file('multiple/models/clip/clip_image_model_vitb32.onnx')
        text_model_path = download_file('multiple/models/clip/clip_text_model_vitb32.onnx')
        self.image_model = Clip._load_model(image_model_path)
        self.text_model = Clip._load_model(text_model_path)
        
        self.embedding_size = 512
        self._tokenizer = Tokenizer()
        self._preprocessor = Preprocessor()
        self._batch_size = batch_size

    @staticmethod
    def _load_model(path: str):
        if not os.path.exists(path):
            s3_url = f"https://lakera-clip.s3.eu-west-1.amazonaws.com/{os.path.basename(path)}"
            logging.info(f"The model file ({path}) doesn't exist or it is invalid. "
                f"Downloading it from the public S3 bucket: {s3_url}.")
            download_url(s3_url, path)
        return ort.InferenceSession(path, providers=ort.get_available_providers())
            
    def get_image_embeddings(self, images: Iterable[Union[Image.Image, np.ndarray]]) -> np.ndarray:
        """Compute the embeddings for a list of images.

        Args:
            images: A list of images to run on. Each image must be a 3-channel
                (RGB) image. Can be any size, as the preprocessing step will
                resize each image to size (224, 224).

        Returns:
            An array of embeddings of shape (len(images), embedding_size).
        """
        embeddings = []
        for batch in to_batches(images, self._batch_size):
            images = [self._preprocessor.encode_image(image) for image in images]
            if not images:
                return self._get_empty_embedding()
            batch = np.concatenate(images)
            embeddings.append(self.image_model.run(None, {"IMAGE": batch})[0])
        if not embeddings:
            return self._get_empty_embedding()
        return np.concatenate(embeddings)

    def get_text_embeddings(self, texts: Iterable[str]) -> np.ndarray:
        """Compute the embeddings for a list of texts.

        Args:
            texts: A list of texts to run on. Each entry can be at most
                77 characters.

        Returns:
            An array of embeddings of shape (len(texts), embedding_size).
        """
        embeddings = []
        for batch in to_batches(texts, self._batch_size):
            text = self._tokenizer.encode_text(batch)
            if len(text) == 0:
                return self._get_empty_embedding()
            embeddings.append(self.text_model.run(None, {"TEXT": text})[0])
        if not embeddings:
            return self._get_empty_embedding()
        return np.concatenate(embeddings)

    def _get_empty_embedding(self):
        return np.empty((0, self.embedding_size), dtype=np.float32)



def to_batches(items, size):
    """
    Splits an iterable (e.g. a list) into batches of length `size`. Includes
    the last, potentially shorter batch.

    Examples:
        >>> list(to_batches([1, 2, 3, 4], size=2))
        [[1, 2], [3, 4]]
        >>> list(to_batches([1, 2, 3, 4, 5], size=2))
        [[1, 2], [3, 4], [5]]

        # To limit the number of batches returned
        # (avoids reading the rest of `items`):
        >>> import itertools
        >>> list(itertools.islice(to_batches([1, 2, 3, 4, 5], size=2), 1))
        [[1, 2]]

    Args:
        items: The iterable to split.
        size: How many elements per batch.
    """
    batch = []
    for item in items:
        batch.append(item)
        if len(batch) == size:
            yield batch
            batch = []
    # The last, potentially incomplete batch
    if batch:
        yield batch


__all__ = [Clip]



from ..client import Abraia

import os
import io
import shutil
import contextlib
import numpy as np

# os.environ['YOLO_VERBOSE'] = 'False'

from ultralytics import YOLO

abraia = Abraia()


def sorted_folders(dir):
    items = [os.path.join(dir, name) for name in os.listdir(dir)]
    sorted_items = sorted(items, key=os.path.getctime)
    return sorted_items


def build_model_name(model_name, task):
    if task == 'segment':
        model_name = f"{model_name}-seg"
    if task == 'classify':
        model_name = f"{model_name}-cls"
    return model_name


class Model:
    def __init__(self, task, model_type='yolov8n', imgsz=640):
        model_name = build_model_name(model_type, task)
        self.model = YOLO(f"{model_name}.pt", verbose=False)
        self.model_name = model_name
        self.metrics = {}
        self.task = task
        self.imgsz = imgsz

    def train(self, project, epochs=100, batch=32, callback=None):
        if callback:
            def on_train_epoch_end(trainer):
                loss_items = trainer.loss_items.cpu().detach().numpy()
                loss = float(np.sum(loss_items)) / len(loss_items)
                acc = trainer.metrics.get('metrics/mAP50(B)', 0) if hasattr(trainer, 'metrics') else 0
                callback({'epoch': trainer.epoch, 'epochs': trainer.epochs, 'loss': loss, 'acc': float(acc)})
            self.model.add_callback('on_train_epoch_end', on_train_epoch_end)
        data = f"{project}" if self.task == 'classify' else f"{project}/data.yaml"
        results = self.model.train(data=data, batch=batch, epochs=epochs, imgsz=self.imgsz)

    def test(self, split='val'):
        out = io.StringIO()
        with contextlib.redirect_stderr(out):
            metrics = self.model.val(split=split)
        return {'mAP': float(metrics.box.map50), 'P': metrics.box.p.tolist(), 'R': metrics.box.r.tolist(), 
                'confusionMatrix': metrics.confusion_matrix.matrix.tolist()}

    def save(self, project, classes, device='cpu', half=False):
        # TODO: Add model name versioning
        out = io.StringIO()
        with contextlib.redirect_stdout(out):
            model_src = self.model.export(format="onnx", device=device, opset=11, half=half)
            shutil.copy(model_src, f"{self.model_name}.onnx")
        abraia.upload_file(f"{self.model_name}.onnx", f"{project}/{self.model_name}.onnx")
        abraia.save_json(f"{project}/{self.model_name}.json", 
                         {'task': self.task, 'inputShape': [1, 3, self.imgsz, self.imgsz], 
                          'classes': classes, 'metrics': self.metrics})

    def run(self, img):
        objects = []
        results = self.model.predict(img, verbose=False)[0]
        if results:
            for k, box in enumerate(results.boxes):
                class_id = int(box.cls)
                label = results.names[class_id]
                score = float(box.conf)
                x1, y1, x2, y2 = box.xyxy.squeeze().tolist()
                x1, y1, x2, y2 = round(x1), round(y1), round(x2), round(y2)
                object = {'label': label, 'score': score, 'box': [x1, y1, x2 - x1, y2 - y1]}
                if self.task == 'segment':
                    object['polygon'] = results.masks[k].xy[0]
                objects.append(object)
        return objects
    
    def compile(self, project, classes, device='hailo8'):
        abraia.download_file(f"{project}/{self.model_name}.onnx", f"{self.model_name}.onnx")
        print("Compile model for edge deployment to hailo hef format...")
        print(f"hailomz compile yolov8n --ckpt yolov8n.onnx --calib-path {project}/train/images --classes {len(classes)} --hw-arch {device} --performance")



import cv2
import numpy as np


def is_clockwise(contour):
    value = 0
    num = len(contour)
    for i in range(num):
        p1 = contour[i]
        p2 = contour[i + 1] if i < num - 1 else contour[0]
        value += (p2[0] - p1[0]) * (p2[1] + p1[1]);
    return value < 0


def get_merge_point_idx(contour1, contour2):                   
    idx1, idx2 = 0, 0
    distance_min = -1
    for i, p1 in enumerate(contour1):
        for j, p2 in enumerate(contour2):
            distance = pow(p2[0] - p1[0], 2) + pow(p2[1] - p1[1], 2);
            if distance_min < 0 or distance < distance_min:
                distance_min = distance
                idx1, idx2 = i, j
    return idx1, idx2


def merge_contours(contour1, contour2, idx1, idx2):
    contour = []
    for i in list(range(0, idx1 + 1)):
        contour.append(contour1[i])
    for i in list(range(idx2, len(contour2))):
        contour.append(contour2[i])
    for i in list(range(0, idx2 + 1)):
        contour.append(contour2[i])
    for i in list(range(idx1, len(contour1))):
        contour.append(contour1[i])
    return np.array(contour)


def merge_with_parent(contour_parent, contour):
    if not is_clockwise(contour_parent):
        contour_parent = contour_parent[::-1]
    if is_clockwise(contour):
        contour = contour[::-1]
    idx1, idx2 = get_merge_point_idx(contour_parent, contour)
    return merge_contours(contour_parent, contour, idx1, idx2)


def approx_contour(contour, approx=0.001):
    epsilon = approx * cv2.arcLength(contour, True)
    contour = cv2.approxPolyDP(contour, epsilon, True)
    return contour.reshape(-1, 2)


def mask_to_polygon(mask, origin=[0, 0], approx=0.001):
    """Returns the largest bounding polygon based on the segmentation mask."""
    contours, hierarchies = cv2.findContours(mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
    contours = [approx_contour(contour, approx) for contour in contours]
    parent_idxs = [int(hierarchy[3]) for hierarchy in hierarchies[0]]
    contours_parent = []
    for contour, parent_idx in zip(contours, parent_idxs):
        contours_parent.append(contour if parent_idx < 0 and len(contour) >= 3 else [])
    for contour, parent_idx in zip(contours, parent_idxs):
        if parent_idx >= 0 and len(contour) >= 3 and len(contours_parent[parent_idx]):
            contours_parent[parent_idx] = merge_with_parent(contours_parent[parent_idx], contour)
    lengths = [len(contour) for contour in contours_parent]
    polygon = contours_parent[np.argmax(lengths)] + np.array(origin)
    return polygon.tolist()


# def mask_to_polygon(mask, origin=[0, 0], approx=0.001):
#     """Returns the largest bounding polygon based on the segmentation mask."""
#     contours = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[0]
#     contours = [approx_contour(contour, approx) for contour in contours]
#     lengths = [len(contour) for contour in contours]
#     polygon = contours[np.argmax(lengths)].reshape(-1, 2)
#     polygon = polygon + np.array(origin)
#     return polygon.tolist()


# def mask_to_polygons(mask):
#     """
#     Convert a binary mask to a list of flattened polygons.

#     Args:
#         mask (np.ndarray): 2D binary mask.

#     Returns:
#         list: List of polygons (as flattened arrays).
#         bool: True if the mask has holes.
#     """
#     mask = np.ascontiguousarray(mask)
#     res = cv2.findContours(mask.astype("uint8"), cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE)
#     hierarchy = res[-1]
#     if hierarchy is None:
#         return [], False
#     has_holes = (hierarchy.reshape(-1, 4)[:, 3] >= 0).sum() > 0
#     contours = res[-2]
#     polygons = [c.flatten() + 0.5 for c in contours if len(c) >= 6]
#     return polygons, has_holes


def approximate_polygon(polygon, approx=0.02):
    contour = np.array([polygon]).astype(np.int32)
    contour = approx_contour(contour, approx)
    return contour.tolist()


def normalize(img, mean, std):
    img = (img / 255 - np.array(mean)) / np.array(std)
    return img.astype(np.float32)


def py_cpu_nms(dets, thresh):
    """Pure Python NMS baseline.

    # --------------------------------------------------------
    # Fast R-CNN
    # Copyright (c) 2015 Microsoft
    # Licensed under The MIT License [see LICENSE for details]
    # Written by Ross Girshick
    # --------------------------------------------------------
    """
    x1, y1 = dets[:, 0], dets[:, 1]
    x2, y2 = dets[:, 2], dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(ovr <= thresh)[0]
        order = order[inds + 1]

    return keep


def nms(dets, thresh):
    """
    Pure Python NMS implementation using NumPy.
    dets: (N, 5) - [x1, y1, x2, y2, score]
    thresh: IoU threshold
    """
    if dets.shape[0] == 0:
        return np.array([], dtype=np.int64)

    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    ndets = dets.shape[0]
    suppressed = np.zeros((ndets), dtype=np.int64)

    for _i in range(ndets):
        i = order[_i]
        if suppressed[i] == 1:
            continue
        ix1 = x1[i]
        iy1 = y1[i]
        ix2 = x2[i]
        iy2 = y2[i]
        iarea = areas[i]
        for _j in range(_i + 1, ndets):
            j = order[_j]
            if suppressed[j] == 1:
                continue
            xx1 = max(ix1, x1[j])
            yy1 = max(iy1, y1[j])
            xx2 = min(ix2, x2[j])
            yy2 = min(iy2, y2[j])
            w = max(0.0, xx2 - xx1 + 1)
            h = max(0.0, yy2 - yy1 + 1)
            inter = w * h
            ovr = inter / (iarea + areas[j] - inter)
            if ovr >= thresh:
                suppressed[j] = 1

    return np.where(suppressed == 0)[0]


# def iou(box1, box2):
#     """Calculates the intersection-over-union of two boxes."""
#     tl1, wh1, br1 = [box1[0], box1[1]], [box1[2], box1[3]], [box1[0] + box1[2], box1[1] + box1[3]]
#     tl2, wh2, br2 = [box2[0], box2[1]], [box2[2], box2[3]], [box2[0] + box2[2], box2[1] + box2[3]]
#     intersection_area = np.prod(np.maximum(np.minimum(br1, br2) - np.maximum(tl1, tl2), 0))
#     union_area = np.prod(wh1) + np.prod(wh2) - intersection_area;
#     return intersection_area / union_area


# def non_maximum_suppression(objects, iou_threshold):
#     results = []
#     objects.sort(key=lambda obj: obj['score'], reverse=True)
#     while len(objects) > 0:
#         results.append(objects[0])
#         objects = [obj for obj in objects if iou(obj['box'], objects[0]['box']) < iou_threshold]
    # return results


def non_maximum_suppression(objects, iou_threshold):
    dets = []
    for obj in objects:
        s = obj['score']
        x, y, w, h = obj['box']
        dets.append([x, y, x + w, y + h, s])
    if dets:
        idxs = py_cpu_nms(np.array(dets), iou_threshold)
        return [objects[idx] for idx in idxs]
    return []


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def softmax(x, axis=-1):
    """Computes softmax values for each sets of scores in x."""
    e_x = np.exp(x - np.max(x, axis=axis, keepdims=True))
    return e_x / e_x.sum(axis=axis, keepdims=True)


def count_objects(results):
    counts = {}
    for result in results:
        label= result['label']
        counts[label] = counts.get(label, 0) + 1
    objects = [{'label': label, 'count': counts[label]} for label in counts.keys()]
    return objects


def triplet_orientation(A, B, C):
    """Return the orientation of the triplet (A, B, C)."""
    val = (C[1] - A[1]) * (B[0] - A[0]) - (B[1] - A[1]) * (C[0] - A[0])
    return 1 if val > 0 else -1 if val < 0 else 0


def segments_intersect(A, B, C, D):
    """Check if line segments AB and CD intersect."""
    o1, o2 = triplet_orientation(A, B, C), triplet_orientation(A, B, D)
    o3, o4 = triplet_orientation(C, D, A), triplet_orientation(C, D, B)
    return (1 if o1 > 0 else -1) if (o1 != o2 and o3 != o4) else 0


def point_in_polygon(point, polygon):
    """Check if a point is inside a polygon using the ray-casting algorithm."""
    x, y = point
    inside = False
    for i in range(len(polygon)):
        (x1, y1), (x2, y2) = polygon[i-1], polygon[i]
        intersect = ((y1 > y) != (y2 > y)) and (x < (x2 - x1) * (y - y1) / (y2 - y1) + x1)
        if intersect:
            inside = not inside
    return inside


def crop_box(img, box):
    x, y, w, h = box
    return img[y:y+w, x:x+w]


def euclidean_distance(feat1, feat2):
    return float(np.linalg.norm(feat1 - feat2))


def cosine_similarity(feat1, feat2):
    """Compute pairwise similarity scores between two vectors."""
    return float(np.dot(feat1, feat2) / (np.linalg.norm(feat1) * np.linalg.norm(feat2)))


def search_vector(vector, index, max_results=1):
    distances = [cosine_similarity(vector, row['vector']) for row in index]
    idxs = np.argsort(distances)[-max_results:][::-1]
    return idxs, [distances[idx] for idx in idxs]


def normalize_vector(vector):
    return vector / np.linalg.norm(vector)



import os
import copy
import time
import onnx
import torch
import torchvision
import numpy as np
import matplotlib.pyplot as plt

from torchvision import models, transforms, datasets

from ..client import Abraia
from ..utils import temporal_src


abraia = Abraia()


torch.backends.cudnn.benchmark = True
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


# def read_image(path):
#     dest = abraia.download_file(path, cache=True)
#     return load_image(dest)


# def load_dataset(dataset, shuffle=True):
#     paths, labels = [], []
#     annotations = abraia.load_json(f"{dataset}/annotations.json")
#     keys = list(filter(lambda k: k != 'filename', annotations[0].keys()))
#     paths = [f"{dataset}/{annotation['filename']}" for annotation in annotations]
#     labels = [annotation['label'] for annotation in annotations]
#     if shuffle:
#         ids = list(range(len(paths)))
#         random.shuffle(ids)
#         paths = [paths[id] for id in ids]
#         labels = [labels[id] for id in ids]
#     return paths, labels


# class Dataset(torch.utils.data.Dataset):
#     def __init__(self, root_dir, transform=None, target_transform=None):
#         paths, labels = load_dataset(root_dir)
#         self.paths = paths
#         self.labels = labels
#         self.root_dir = root_dir
#         self.transform = transform
#         self.target_transform = target_transform
#         self.classes = list(np.sort(np.unique(labels)))

#     def __len__(self):
#         return len(self.paths)

#     def __getitem__(self, idx):
#         if torch.is_tensor(idx):
#             idx = idx.tolist()
#         path = self.paths[idx]
#         label = self.labels[idx]
#         label = self.classes.index(label)
#         img = read_image(path)
#         if self.transform:
#             img = self.transform(img)
#         if self.target_transform:
#             label = self.target_transform(label)
#         return img, label


def create_model(class_names, pretrained=True):
    model = models.resnet18(weights=models.ResNet18_Weights.IMAGENET1K_V1 if pretrained else None)
    for param in model.parameters():
        param.requires_grad = False
    num_ftrs = model.fc.in_features
    model.fc = torch.nn.Linear(num_ftrs, len(class_names))
    model.to(device)
    return model


transform = transforms.Compose([
    transforms.Resize(256),
    transforms.CenterCrop(224),
    transforms.ToTensor(),
    transforms.Normalize(
        mean=[0.485, 0.456, 0.406],
        std=[0.229, 0.224, 0.225])
])


# License: BSD
# Author: Sasank Chilamkurthy

def train_model(model, dataloaders, criterion=None, optimizer=None, scheduler=None, num_epochs=25, callback=None):
    criterion = criterion or torch.nn.CrossEntropyLoss()
    optimizer = optimizer or torch.optim.SGD(model.fc.parameters(), lr=0.001, momentum=0.9)
    scheduler = scheduler or torch.optim.lr_scheduler.StepLR(optimizer, step_size=7, gamma=0.1)
    best_model_wts = copy.deepcopy(model.state_dict())
    best_acc = 0.0
    for epoch in range(num_epochs):
        print(f'Epoch {epoch}/{num_epochs - 1}')
        print('-' * 10)
        # Each epoch has a training and validation phase
        for phase in ['train', 'val']:
            if phase == 'train':
                model.train()  # Set model to training mode
            else:
                model.eval()   # Set model to evaluate mode
            running_loss = 0.0
            running_corrects = 0
            # Iterate over data
            for inputs, labels in dataloaders[phase]:
                inputs = inputs.to(device)
                labels = labels.to(device)
                # zero the parameter gradients
                optimizer.zero_grad()
                # forward
                # track history if only in train
                with torch.set_grad_enabled(phase == 'train'):
                    outputs = model(inputs)
                    _, preds = torch.max(outputs, 1)
                    loss = criterion(outputs, labels)
                    # backward + optimize only if in training phase
                    if phase == 'train':
                        loss.backward()
                        optimizer.step()
                # statistics
                running_loss += loss.item() * inputs.size(0)
                running_corrects += torch.sum(preds == labels.data)
            if phase == 'train':
                scheduler.step()
            epoch_loss = running_loss / len(dataloaders[phase].dataset)
            epoch_acc = running_corrects.double() / len(dataloaders[phase].dataset)
            print(f'{phase} Loss: {epoch_loss:.4f} Acc: {epoch_acc:.4f}')
            if callback:
                callback({'epoch': epoch, 'epochs': num_epochs, 'loss': epoch_loss, 'acc': float(epoch_acc)})
            # deep copy the model
            if phase == 'val' and epoch_acc > best_acc:
                best_acc = epoch_acc
                best_model_wts = copy.deepcopy(model.state_dict())
        print()
    print(f'Best val Acc: {best_acc:4f}')
    # load best model weights
    model.load_state_dict(best_model_wts)
    return model


def imshow(inp, title=None):
    """Imshow for Tensor."""
    inp = inp.numpy().transpose((1, 2, 0))
    mean = np.array([0.485, 0.456, 0.406])
    std = np.array([0.229, 0.224, 0.225])
    inp = std * inp + mean
    inp = np.clip(inp, 0, 1)
    plt.imshow(inp)
    if title is not None:
        plt.title(title)
    plt.pause(0.001)  # pause a bit so that plots are updated


def visualize_data(dataloader):
    class_names = dataloader.dataset.classes
    inputs, classes = next(iter(dataloader))
    out = torchvision.utils.make_grid(inputs)  # Make a grid from batch
    imshow(out, title=[class_names[x] for x in classes])


def visualize_model(model, dataloader, num_images=6):
    class_names = dataloader.dataset.classes
    was_training = model.training
    model.eval()
    images_so_far = 0
    fig = plt.figure()
    with torch.no_grad():
        for i, (inputs, labels) in enumerate(dataloader):
            inputs = inputs.to(device)
            labels = labels.to(device)
            outputs = model(inputs)
            _, preds = torch.max(outputs, 1)
            for j in range(inputs.size()[0]):
                images_so_far += 1
                ax = plt.subplot(num_images//2, 2, images_so_far)
                ax.axis('off')
                ax.set_title(f'predicted: {class_names[preds[j]]}')
                imshow(inputs.cpu().data[j])
                if images_so_far == num_images:
                    model.train(mode=was_training)
                    return
        model.train(mode=was_training)


class Model:
    def __init__(self):
        self.imgsz = 224
        self.input_shape = [1, 3, self.imgsz, self.imgsz]
        self.model_name = 'resnet18'

    def create_dataset(self, dataset, batch=8):
        # Data augmentation and normalization for training
        # Just normalization for validation
        data_transforms = {
            'train': transforms.Compose([
                transforms.RandomResizedCrop(self.imgsz),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
            ]),
            'val': transforms.Compose([
                transforms.Resize(256),
                transforms.CenterCrop(self.imgsz),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
            ]),
        }
        image_datasets = {x: datasets.ImageFolder(os.path.join(dataset, x), transform=data_transforms[x]) for x in ['train', 'val']}
        dataloaders = {x: torch.utils.data.DataLoader(image_datasets[x], batch_size=batch, shuffle=True, num_workers=4) for x in ['train', 'val']}
        classes = image_datasets['train'].classes
        return dataloaders, classes

    def train(self, dataset, epochs=25, batch=8, callback=None):
        dataloaders, classes = self.create_dataset(dataset, batch=batch)
        model_conv = create_model(classes)
        self.dataloaders = dataloaders
        self.classes = classes
        since = time.time()
        self.model = train_model(model_conv, dataloaders, num_epochs=epochs, callback=callback)
        time_elapsed = time.time() - since
        print(f'Training complete in {time_elapsed // 60:.0f}m {time_elapsed % 60:.0f}s')

    def test(self, split='val'):
        self.model.eval()
        running_corrects = 0
        confusion_matrix = np.zeros((len(self.classes), len(self.classes)), dtype=int)
        with torch.no_grad():
            for inputs, labels in self.dataloaders[split]:
                inputs = inputs.to(device)
                labels = labels.to(device)
                outputs = self.model(inputs)
                _, preds = torch.max(outputs, 1)
                running_corrects += torch.sum(preds == labels.data)
                for t, p in zip(labels.view(-1), preds.view(-1)):
                    confusion_matrix[t.long(), p.long()] += 1
        acc = running_corrects.double() / len(self.dataloaders[split].dataset)
        return {'acc': float(acc), 'confusionMatrix': confusion_matrix.tolist()}

    def save(self, dataset, classes, device='cpu'):
        self.model.to(device)
        model_src = temporal_src(f"{dataset}/{self.model_name}.onnx")
        dummy_input = torch.randn(1, 3, self.imgsz, self.imgsz)
        torch.onnx.export(self.model, dummy_input, model_src, export_params=True, opset_version=10, do_constant_folding=True, input_names=['input'], output_names=['output'])
        onnx_model = onnx.load(model_src)
        onnx.checker.check_model(onnx_model)
        onnx.save(onnx_model, model_src)
        abraia.upload_file(model_src, f"{dataset}/{self.model_name}.onnx")
        abraia.save_json(f"{dataset}/{self.model_name}.json", {'inputShape': self.input_shape, 'classes': classes})

    def run(self, img):
        input_tensor = transform(img)
        input_batch = input_tensor.unsqueeze(0)
        output = self.model(input_batch)
        pred = torch.softmax(output.squeeze(0), dim=0)
        idx = int(pred.argmax())
        score = float(pred[idx])
        label = self.classes[idx]
        return [{'label': label, 'score': score}]



import os
import time

from tqdm import tqdm
from glob import glob

from abraia.inference import Model, Tracker, FaceRecognizer, FaceAttribute, PlateRecognizer
from abraia.inference.faces import find_pose
from abraia.inference.clip import Clip
from abraia.inference.ops import count_objects, search_vector
from abraia.inference.tools import LineCounter, RegionFilter, RegionTimer
from abraia.utils.draw import render_results, render_counter, render_region, draw_overlay, draw_text_multiline
from abraia.utils import Video, download_url, load_image, show_image


DEMOS = {
    'tomato': {
        'model': 'multiple/tomato/yolov8n_v6.onnx',
        'src': '10179855-hd_1280_720_30fps.mp4',
        'labels': ['tomato'],
        'counter': [(960, 0), (960, 720)],
        'dest': 'tomato_onnx.mp4'
    },
    'apple': {
        'model': 'multiple/models/yolov8n-seg.onnx',
        'src': '5479199-hd_1280_720_25fps.mp4',
        'labels': ['apple'],
        'dest': 'apple_onnx.mp4'
    },
    'strawberry': {
        'model': 'multiple/strawberry/yolov8n.onnx',
        'src': '9710983-hd_1920_1080_30fps.mp4',
        'labels': ['strawberry']
    },
    'grapes': {
        'model': 'multiple/grapes/yolov8n.onnx',
        'src': '5658544-hd_1366_720_24fps.mp4',
        'labels': ['grapes']
    },
    'people': {
        'src': '853889-hd_1920_1080_25fps.mp4',
        'labels': ['person'],
        'counter': [(0, 650), (1920, 650)],
        'region': [(0, 600), (1920, 600), (1920, 700), (0, 700)]
    },
    'queue': {
        'src': '4775505-hd_1920_1080_30fps.mp4',
        'labels': ['person'],
        'timer': [(10, 600), (1690, 600), (1690, 700), (10, 700)]
    },
    'escalator': {
        'src': '14393755-hd_1920_1080_30fps.mp4',
        'labels': ['person'],
        'counter': [(950, 670), (270, 895)],
        'region': [[0, 245], [350, 1080], [1200, 1080], [530, 0], [0, 0]],
        'timer': [[0, 245], [350, 1080], [1200, 1080], [530, 0], [0, 0]]
    }
}


HAILO_DEMOS = {
    'tomato': {
        'hef_path': 'multiple/tomato/yolov8n.hef',
        'src': '10179855-hd_1280_720_30fps.mp4',
        'dest': 'tomato_hef.mp4'
    },
    'apple': {
        'hef_path': 'yolov5m_seg_with_nms',
        'task': 'segment',
        'src': '5479199-hd_1280_720_25fps.mp4',
        'dest': 'apple_hef.mp4'
    },
    'segment': {
        'hef_path': 'yolov5m_seg_with_nms',
        'task': 'segment',
        'src': '853889-hd_1920_1080_25fps.mp4'
    },
    'pose': {
        'hef_path': 'yolov8m_pose',
        'task': 'pose',
    }
}


def monitor_objects(src=None, demo='detect', resolution=(1280, 720)):
    """Monitor, count, or just detect objects in a video stream."""
    print(f"Available demos: {', '.join(DEMOS.keys())}")
    selected = DEMOS.get(demo) or {}
    src = src or selected.get('src', 0)
    if isinstance(src, str) and not os.path.exists(src) and src.endswith('.mp4'):
        download_url(f"https://api.abraia.me/files/multiple/videos/{src}", src)

    video = Video(src, resolution=resolution, dest=selected.get('dest'))
    tracker = Tracker(frame_rate=video.frame_rate)
    model = Model(selected.get('model', 'multiple/models/yolov8n.onnx'))
    
    line_counter = LineCounter(selected['counter']) if selected.get('counter') else None
    region_filter = RegionFilter(selected['region']) if selected.get('region') else None
    region_timer = RegionTimer(selected['timer']) if selected.get('timer') else None

    labels = selected.get('labels')
    for k, frame in enumerate(video):
        frame_time = round(k / video.frame_rate, 2)
        t0 = time.time()
        results = model.run(frame, labels=labels)
        if region_filter:
            results, _ = region_filter.update(results)
        results = tracker.update(results)
        out = frame.copy()
        if line_counter:
            in_count, out_count = line_counter.update(results)
            out = render_counter(out, line_counter.line, f"In: {in_count} | Out: {out_count}")
        if region_timer:
            in_objects, out_objects = region_timer.update(results, frame_time)
            out = render_region(out, region_timer.region, f"Count: {len(in_objects)}", color=(255, 255, 0))
            results = in_objects
        out = render_results(out, results)
        print(f"#{k} {round((time.time() - t0) * 1000, 1)}ms {count_objects(results)}")
        video.show(out)

    if line_counter:
        print(f"Final In: {in_count}, Final Out: {out_count}")


def monitor_objects_hailo(src=None, demo='detect'):
    """Monitor, count, or just detect objects in a video stream using Hailo."""
    from abraia.hailo import detect
    print(f"Available Hailo demos: {', '.join(HAILO_DEMOS.keys())}")
    selected = HAILO_DEMOS.get(demo) or {}
    src = src or selected.get('src', 0)
    if isinstance(src, str) and not os.path.exists(src) and src.endswith('.mp4'):
        download_url(f"https://api.abraia.me/files/multiple/videos/{src}", src)
    options = selected.copy()
    options['input'] = src
    detect.main(**options)


def track_faces(src=None, resolution=(1280, 720)):
    """Track faces in a video stream from a file or webcam."""
    recognition = FaceRecognizer()
    attribute = FaceAttribute()
    index = []
    src = src or 0
    video = Video(src, resolution=resolution)
    for frame in video:
        results = recognition.detect_faces(frame)
        faces = recognition.extract_faces(frame, results)
        results = recognition.identify_faces(frame, results, index)
        for k, (face, result) in enumerate(zip(faces, results)):
            (h, w), (fh, fw) = frame.shape[:2], face.shape[:2]
            frame = draw_overlay(frame, face, [w - fw, k * fh, fw, fh])
            roll, yaw, pitch = find_pose(result['keypoints'])
            draw_text_multiline(frame, [f"Roll: {roll} degrees",
                                        f"Pitch: {pitch} degrees",
                                        f"Yaw: {yaw} degrees"], (10, 140), text_color=(255, 0, 0), text_scale=0.6)
            if result['label'] == 'unknown':
                index.append({'name': f"face_{len(index)}", 'vector': result['vector']})
            gender, age, score = attribute.predict(face)
            print(f"{gender[0]} {age}, {score}")
            result['label'] = f"{gender[0]} {age} ({result['label']})"
        frame = render_results(frame, results)
        video.show(frame)


def detect_plates(src=None):
    """Detect license plates in a video and show OCR results."""
    src = src or 'cars.mp4'
    if isinstance(src, str) and not os.path.exists(src) and src.endswith('.mp4'):
        download_url(f"https://api.abraia.me/files/multiple/videos/{src}", src)
    video = Video(src)
    recognizer = PlateRecognizer()
    for frame in video:
        results = recognizer.recognize(frame)
        frame = render_results(frame, results)
        video.show(frame)


def retrieve_images(query="man with red shirt", path='images/*.jpg'):
    """Search for images using a text query."""
    clip_model = Clip()
    print("Building image index...")
    image_paths = glob(path)
    image_index = [{'vector': clip_model.get_image_embeddings([load_image(image_path)])[0]} for image_path in tqdm(image_paths)]
    print("Searching for images...")
    print("Query:", query)
    vector = clip_model.get_text_embeddings([query])[0]
    idxs, scores = search_vector(vector, image_index, max_results=5)
    for idx in idxs:
        img = load_image(image_paths[idx])
        show_image(img)



import os
import shutil
import itertools

from PIL import Image
from typing import Dict, Any
from tqdm.contrib.concurrent import process_map

from ..utils import save_text
from .dataset import list_datasets, load_dataset, search_images, list_models, download_file


def load_tasks(task):
    if task:
        tasks = ['classify', 'detect', 'segment']
        return tasks[:tasks.index(task)+1]
    return []


def save_annotation(annotation, folder, classes, task):
    im = Image.open(os.path.join(folder, 'images', annotation['filename']))
    label_lines = []
    for object in annotation.get('objects', []):
        label, box, polygon = object.get('label'), object.get('box'), object.get('polygon')
        if task == 'segment':
            if polygon:
                label_line = f"{classes.index(label)} " + ' '.join([f"{point[0] / im.width} {point[1] / im.height}" for point in polygon])
                label_lines.append(label_line)
        elif task == 'detect':
            if polygon:
                xx, yy = [point[0] for point in polygon], [point[1] for point in polygon]
                x1, y1, x2, y2 = min(xx), min(yy), max(xx), max(yy)
                box = [x1, y1, x2 - x1, y2 - y1]
            if box:
                label_line = f"{classes.index(label)} {(box[0] + box[2] / 2) / im.width} {(box[1] + box[3] / 2) / im.height} {box[2] / im.width} {box[3] / im.height}"
                label_lines.append(label_line)
    label_path = os.path.join(folder, 'labels',  f"{os.path.splitext(annotation['filename'])[0]}.txt")
    save_text(label_path, '\n'.join(label_lines))


def save_data(annotation, folder, classes, task):
    path = annotation['path']
    dest = folder if task == 'classify' else os.path.join(folder, 'images')
    if task == 'classify':
        label = next((obj.get('label', '') for obj in annotation.get('objects', [])), '')
        dest = os.path.join(dest, label)
    download_file(path, dest)
    if task != 'classify':
        save_annotation(annotation, folder, classes, task)


def save_config(dataset, classes):
    yaml_content = f'''
    train: {os.path.join(os.getcwd(), dataset, 'train/images')}
    val: {os.path.join(os.getcwd(), dataset, 'val/images')}
    test: {os.path.join(os.getcwd(), dataset, 'test/images')}
    names: {classes}
    '''
    path = os.path.join(dataset, 'data.yaml')
    save_text(path, yaml_content)

    
def prepare_dataset(dataset, force = False):
    if force or not os.path.exists(dataset.project):
        splits = list(zip(['train', 'val', 'test'], dataset.split()))
        all_annotations, all_folders = [], []
        for x, annotations in splits:
            folder = os.path.join(dataset.project, x)
            all_annotations.extend(annotations)
            all_folders.extend([folder] * len(annotations))
        process_map(save_data, all_annotations, all_folders, itertools.repeat(dataset.classes), itertools.repeat(dataset.task), max_workers=5, chunksize=1, desc="Downloading images")
        if dataset.task != 'classify':
            save_config(dataset.project, dataset.classes)


class ModelTrainer:
    """High-level trainer orchestrator using models and dataset utilities."""
    def __init__(self, project: str, task: str, classes: list, imgsz: int = None):
        self.project = project
        self.task = task
        self.classes = classes
        imgsz = imgsz or (224 if task == 'classify' else 640)
        if task == 'classify':
            from . import classify
            self.model = classify.Model()
        else:
            from . import detect
            self.model = detect.Model(task, imgsz=imgsz)

    def train(self, epochs: int = None, batch: int = 32, callback=None) -> None:
        epochs = epochs or (30 if self.task == 'classify' else 300)
        self.model.train(self.project, epochs=epochs, batch=batch, callback=callback)

    def test(self, split: str = 'val') -> Dict[str, Any]:
        return self.model.test(split=split)

    def save(self, device='cpu') -> None:
        self.model.save(self.project, self.classes, device=device)

    def run(self, img):
        return self.model.run(img)

    def compile(self, device='hailo8'):
        if self.task != 'detect':
            raise NotImplementedError("Model compilation is only implemented for detection models.")
        self.model.compile(self.project, self.classes, device=device)
    


import os
import re
import io
import urllib
import requests
import filetype
import itertools

from tqdm import tqdm
from PIL import Image

from ..client import Abraia
from ..utils import HEADERS, load_image, load_url, list_dir, url_path
from ..inference.detect import segment_objects


abraia = Abraia()

GOOGLE_BASE_URL = 'https://www.google.com/search?q='
GOOGLE_PICTURE_ID = '''&biw=1536&bih=674&tbm=isch&sxsrf=ACYBGNSXXpS6YmAKUiLKKBs6xWb4uUY5gA:1581168823770&source=lnms&sa=X&ved=0ahUKEwioj8jwiMLnAhW9AhAIHbXTBMMQ_AUI3QUoAQ'''

BING_BASE_URL = 'https://www.bing.com/images/async?q='


def convert_to_jpg(src, save_output, max_size=1920):
    import imagehash
    im = Image.open(src).convert('RGB')
    im.thumbnail([max_size, max_size], Image.LANCZOS)
    phash = str(imagehash.phash(im))
    im.save(os.path.join(save_output, phash + '.jpg'))


def download_page(url):
    resp = requests.get(url, headers=HEADERS)
    resp.raise_for_status()
    return resp.text


def save_image_file(link, save_output, timeout=10, max_size=1920):
    resp = requests.get(link, headers=HEADERS, allow_redirects=True, timeout=timeout)
    kind = filetype.guess(resp.content)
    if kind and kind.mime.startswith('image'):
        d = io.BytesIO(resp.content)
        convert_to_jpg(d, save_output, max_size)
    else:
        raise ValueError(f'Invalid image, not saving')


def scan_bing_page(html):
    links = re.findall('murl&quot;:&quot;(.*?)&quot;', html)
    for link in links:
        link = link.replace(" ", "%20")
        yield link


def search_bing(query, limit=50):
    for page_counter in range(100):
        request_url = BING_BASE_URL + urllib.parse.quote_plus(query) \
                        + '&first=' + str(page_counter) + '&count=' + str(limit) + '&adlt=off'
        html = download_page(request_url)
        for link in scan_bing_page(html):
            yield link


def scan_google_page(html, extensions={'.jpg', '.jpeg', '.webp'}, timer=5000):
    """Scans for pictures to download based on the keywords"""
    SCANNER_COUNTER = -1
    scanner = html.find
    for _ in range(timer):
        new_line = scanner('"https://', SCANNER_COUNTER + 1)  # How Many New lines
        SCANNER_COUNTER = scanner('"', new_line + 1)  # Ends of line
        buffor = scanner('\\', new_line + 1, SCANNER_COUNTER)
        last_line = buffor if buffor != -1 else SCANNER_COUNTER
        link = html[new_line + 1:last_line]
        if any(extension in link for extension in extensions):
            link = link.replace(" ", "%20")
            yield link


def search_google(query):
    request_url = GOOGLE_BASE_URL + urllib.parse.quote_plus(query) + GOOGLE_PICTURE_ID
    html = download_page(request_url)
    for link in scan_google_page(html):
        yield link


def download(query, limit=100, save_output='dataset', verbose=True):
    seen = set()
    download_count = 0
    os.makedirs(save_output, exist_ok=True)
    links = [search_google(query), search_bing(query)]
    ends = [False] * len(links)
    for id in itertools.cycle(range(len(links))):
        try:
            link = next(links[id])
            if link not in seen:
                seen.add(link)
                if download_count < limit:
                    try:
                        save_image_file(link, save_output)
                        download_count += 1
                        if verbose:
                            print(f"[%] Downloaded Image #{download_count} from {link}")
                    except Exception as e:
                        print(f"[!] Error getting {link}: {e}")
                else:
                    break
        except StopIteration:
            ends[id] = True
            if set(ends) == {True}:
                break


def search_images(query, limit=100, save_output='dataset', verbose=True):
    """Search and download images from Google and Bing."""
    download(query, limit=limit, save_output=save_output,  verbose=verbose)
    return list_dir(save_output)


def download_file(path, folder):
    dest = os.path.join(folder, os.path.basename(path))
    if not os.path.exists(dest):
        abraia.download_file(path, dest)
    return dest


def detect_dino(img, classes, threshold=0.3, pipe=None):
    """Detect objects in an image using Grounding Dino."""

    classes = [label.lower().strip() for label in classes]
    labels = [f"{label}." if not label.endswith('.') else label for label in classes]
    if pipe is None:
        from transformers import pipeline
        pipe = pipeline(task="zero-shot-object-detection", model="IDEA-Research/grounding-dino-tiny")
    results = pipe(Image.fromarray(img), candidate_labels=labels, threshold=threshold)

    objects = []
    for result in results:
        score = result["score"]
        if score > threshold:
            label = result["label"].rpartition('.')[0]
            xmin, ymin, xmax, ymax = result['box'].values()
            objects.append({"label": label, "score": score, "box": [xmin, ymin, xmax - xmin, ymax - ymin]})
    return objects


def list_datasets():
    folders = abraia.list_files()[1]
    return [folder['name'] for folder in folders if abraia.check_file(f"{folder['name']}/annotations.json")]


def list_models(project):
    files = abraia.list_files(f"{project}/")[0]
    return [f['name'] for f in files if f['name'].endswith('.onnx')]


def load_annotations(project):
    annotations = abraia.load_json(f"{project}/annotations.json")
    for annotation in annotations:
        annotation['path'] = f"{project}/{annotation['filename']}"
        annotation['url'] = url_path(f"{abraia.userid}/{annotation['path']}")
    return annotations


def load_labels(annotations):
    labels = []
    for annotation in annotations:
        for object in annotation.get('objects', []):
            label = object.get('label')
            if label and label not in labels:
                labels.append(label)
    return list(set(labels))


def load_task(annotations):
    classify, detect, segment = False, False, False
    for annotation in annotations:
        for object in annotation.get('objects', []):
            if 'polygon' in object:
                segment = True
            elif 'box' in object:
                detect = True
            elif 'label' in object:
                classify = True
    return 'segment' if segment else 'detect' if detect else 'classify' if classify else ''


def list_images(project):
    files = abraia.list_files(f"{project}/")[0]
    files = [f for f in files if f['type'] in ['image/jpeg', 'image/png']]
    for data in files:
        data['url'] = url_path(f"{abraia.userid}/{data['path']}")
    return files


def save_annotations(project, annotations):
    abraia.save_json(f"{project}/annotations.json", annotations)


class Dataset:
    def __init__(self, project):
        self.project = project
        self.annotations = []
        self.classes = []
        self.task = ''
        self.images = []

    def load(self):
        if self.project in list_datasets():
            self.annotations = load_annotations(self.project)
            self.classes = load_labels(self.annotations)
            self.task = load_task(self.annotations)
            self.images = list_images(self.project)
        return self
    
    def annotate(self, label, segment=False, callback=None):
        from transformers import pipeline
        annotated_filenames = {a['filename'] for a in self.annotations}
        images = [img for img in self.images if img['name'] not in annotated_filenames]
        pipe = pipeline(task="zero-shot-object-detection", model="IDEA-Research/grounding-dino-tiny")
        for i, row in enumerate(tqdm(images)):
            url, filename = row['url'], row['name']
            img = load_image(load_url(url))
            objects = detect_dino(img, [label], pipe=pipe)
            if callback:
                callback({'current': i + 1, 'total': len(images)})
            if objects:
                if segment:
                    try:
                        objects = segment_objects(img, objects)
                    except:
                        continue
                annotation = {'url': url, 'filename': filename, 'objects': objects}
                self.annotations.append(annotation)
        return self.annotations

    def save(self, annotations=None):
        if annotations is not None:
            self.annotations = annotations
        save_annotations(self.project, self.annotations)

    def split(self):
        from sklearn.model_selection import train_test_split
        # TODO: Split dataset by classes to avoid class imbalance
        backgrounds = [annotation for annotation in self.annotations if not annotation.get('objects')]
        annotations = [annotation for annotation in self.annotations if annotation.get('objects')]
        train, test = train_test_split(annotations, test_size=0.3)
        val, test = train_test_split(test, test_size=0.5)
        train.extend(backgrounds)
        return train, val, test

        
def load_dataset(project):
    return Dataset(project).load()
